#include <cstdlib>
#include <cstdio>
#include <cmath>

#include "../coco/coco.h"
#include "indicator.h"
#include "population.h"

Population::Population(int size, int dimension, int n_objectives) :
  size_(size), current_size_(0), dimension_(dimension), n_objectives_(n_objectives)
{
  population_ = (double **) coco_allocate_memory(size_*sizeof(double *));
  for (int i=0; i<size_; i++){
    population_[i] = (double *) coco_allocate_memory(dimension_*sizeof(double));
  }

  values_ = (double **) coco_allocate_memory(size_*sizeof(double *));
  for (int i=0; i<size_; i++){
    values_[i] = (double *) coco_allocate_memory(n_objectives_*sizeof(double));
  }

  indicator_values_ = (double **) coco_allocate_memory(size_*sizeof(double *));
  for (int i=0; i<size_; i++){
    indicator_values_[i] = (double *) coco_allocate_memory(size_*sizeof(double));
  }

  fitnesses_ = (double *) coco_allocate_memory(size_*sizeof(double));

  bmin_ = (double *) coco_allocate_memory(size_*sizeof(double));

  bmax_ = (double *) coco_allocate_memory(size_*sizeof(double));
}

Population::~Population(){
  /* Free populations */
  for (int i=0; i<size_; i++){
    coco_free_memory(population_[i]);
  }
  coco_free_memory(population_);

  /* Free values */
  for (int i=0; i<size_; i++){
    coco_free_memory(values_[i]);
  }
  coco_free_memory(values_);

  /* Free indicator values */
  for (int i=1; i<size_; i++){
    coco_free_memory(indicator_values_[i]);
  }
  coco_free_memory(indicator_values_);

  /* Free fitnesses */
  coco_free_memory(fitnesses_);

  coco_free_memory(bmin_);

  coco_free_memory(bmax_);
}

int Population::getSize(){
  return size_;
}

int Population::getCurrentSize(){
  return current_size_;
}

void Population::setCurrenSize(int n){
  current_size_ = n;
}

int Population::getDim(){
  return dimension_;
}

void Population::empty(){
  current_size_ = 0;
}

void Population::setIndiv(int i, int j, double value){
  population_[i][j] = value;
}

double Population::getIndiv(int i, int j){
  return population_[i][j];
}

void Population::setFit(int i, double value){
  fitnesses_[i] = value;
}

double Population::getFit(int i){
  return fitnesses_[i];
}

void Population::setMin(int i){
  ind_min_ = i;
}

int Population::getMin(){
  return ind_min_;
}

void Population::random_init(const double *lower_bounds,
                             const double *upper_bounds,
                             coco_random_state_t *random_generator)
{
  double range;
  for (int i=0; i < size_; i++){
    for (int j=0; j < dimension_; j++) {
      range = upper_bounds[j] - lower_bounds[j];
      population_[i][j] = lower_bounds[j] + coco_random_uniform(random_generator) * range;
    }
  }
  current_size_ = size_;
}

void Population::evaluate(evaluate_function_t evaluate_function){
  for (int i=0; i<current_size_; i++){
    evaluate_function(population_[i], values_[i]);
  }
}

void Population::compute_indicators(){

  double tmp_bmin, tmp_bmax;

  for (int i = 0; i < n_objectives_; i++) {
    for (int j = 0; j < current_size_; j++ ) {
      if (j==0) {
        tmp_bmin = values_[j][i];
        tmp_bmax = values_[j][i];
      }
      else {
        if (values_[j][i]>tmp_bmax) {
          tmp_bmax = values_[j][i];
        }
        if (values_ [j][i]<tmp_bmin) {
          tmp_bmin = values_[j][i];
        }
      }
    }
    bmin_[i] = tmp_bmin;
    bmax_[i] = tmp_bmax;
  }

  c_ = 0;
  for (int i=0; i<current_size_; i++){
    for (int j=0; j<current_size_; j++){
      indicator_values_[i][j] = epsilon_indicator(values_[i], values_[j], bmin_, bmax_, n_objectives_);
      if (abs(indicator_values_[i][j]) > c_){
        c_ = abs(indicator_values_[i][j]);
      }
    }
  }
}

void Population::compute_fitnesses(double kappa){
  double d = c_ * kappa;
  double fitness;
  int min;

  for (int i=0; i<current_size_; i++){
    fitness = 0;
    for (int j=0; j<current_size_; j++){
      if (i != j){
        fitness += -exp(-indicator_values_[j][i]/d);
      }
    }
    fitnesses_[i] = fitness;
    if (i == 0) {
      min = 0;
    }
    else {
      if (fitnesses_[min] > fitnesses_[i]) {
        min = i;
      }
    }
  }
  ind_min_ = min;
}

void Population::darwin(double kappa){
  double d = c_ * kappa;
  double * tmp_p;
  double tmp;
  int min;

  /* Update fitnesses */
  for (int i=0; i<current_size_; i++){
    if (i != ind_min_){
      fitnesses_[i] += exp(-indicator_values_[ind_min_][i]/d);
    }
  }

  /* Remove the individu with the minimum fitness : swapp values with last individu */
  tmp_p = population_[ind_min_];
  population_[ind_min_] = population_[current_size_-1];
  population_[current_size_-1] = tmp_p;

  tmp_p = values_[ind_min_];
  values_[ind_min_] = values_[current_size_-1];
  values_[current_size_-1] = tmp_p;

  tmp_p = indicator_values_[ind_min_];
  indicator_values_[ind_min_] = indicator_values_[current_size_-1];
  indicator_values_[current_size_-1] = tmp_p;
  for (int j=0; j<current_size_; j++){
    tmp = indicator_values_[j][ind_min_];
    indicator_values_[j][ind_min_] = indicator_values_[j][current_size_-1];
    indicator_values_[j][current_size_-1] = tmp;
  }

  tmp = fitnesses_[ind_min_];
  fitnesses_[ind_min_] = fitnesses_[current_size_-1];
  fitnesses_[current_size_-1] = tmp;

  /* Update current size */
  current_size_ -= 1;

  /* Update min indice */
  for (int i=0; i < current_size_; i++){
    if (i == 0){
      min = 0;
    }
    else {
      if (fitnesses_[min] > fitnesses_[i]){
        min = i;
      }
    }
  }
  ind_min_ = min;

  /* Update c_ = max(abs(indicator_values))*/
  c_ = 0;
  for (int i=0; i < current_size_; i++){
    for (int j=i+1; j < current_size_; j++){
      if (abs(indicator_values_[i][j]) > c_){
        c_ = abs(indicator_values_[i][j]);
      }
    }
  }
}

void Population::printPopulation(){
  printf("Population : \n");
  for (int i=0; i < size_; i++){
    for (int j=0; j < dimension_; j++){
      printf("%+6.4lf ", population_[i][j]);
    }
    printf("\n");
  }
}

void Population::printValues(){
  printf("Values : \n");
  for (int i=0; i < size_; i++){
    for (int j=0; j < dimension_; j++){
      printf("%+6.4lf ", values_[i][j]);
    }
    printf("\n");
  }
}

void Population::printIndicatorValues(){
  printf("Indicator Values : \n");
  for (int i=0; i < size_; i++){
    for (int j=0; j < size_; j++){
      printf("%+6.4lf ", indicator_values_[i][j]);
    }
    printf("\n");
  }
}

void Population::printFitnesses(){
  printf("Fitnesses : \n");
  for (int i=0; i < size_; i++){
    printf("%+6.4lf ", fitnesses_[i]);
  }
  printf("\n");
}
